from protocolinterface import ProtocolInterface, val
from playmolecule.utils import get_destination_folder, ensurelist, _requestURL
import json
import os
import shutil
import tempfile
import time
import logging
import enum


logger = logging.getLogger(__name__)


class JobBucketNotFoundError(Exception):
    pass


class JobConfigNotLoadedError(Exception):
    pass


@enum.unique
class JobStatus(enum.IntEnum):
    """Job status codes describing the current status of a job

    WAITING_DATA : Waiting for communication fromm the job. Job has not reached the queue yet for computation.
    VALIDATING : Job is pending validation by admins
    WAITING_APPROVAL : Job is pending approval by admins
    RUNNING : Job is currently running
    COMPLETED : Job has successfully completed
    ERROR : Job has exited with an error
    QUEUED : Job is queued for execution
    SLEEPING : Job is sleeping and will be respawned soon. Used by periodically executing apps.
    """

    WAITING_DATA = 0
    VALIDATING = 1
    WAITING_APPROVAL = 2
    RUNNING = 3
    COMPLETED = 4
    ERROR = 5
    QUEUED = 6
    SLEEPING = 7

    def describe(self):
        codes = {
            0: "Waiting data",
            1: "Validating",
            2: "Waiting approval",
            3: "Running",
            4: "Completed",
            5: "Error",
            6: "Queued",
            7: "Sleeping",
        }
        return codes[self.value]

    def __str__(self):
        return self.describe()


class JobError(Exception):
    pass


def _throwError(message, _logger=False):
    if _logger:
        logger.error(message)
    else:
        raise JobError(message)


def _request_url(*args, **kwargs):
    return _requestURL(*args, **kwargs, _throwError=_throwError)


def short_execid(execid):
    """Returns the short execution ID of a job from the full execution ID"""
    return execid[:8].upper()


class Job(ProtocolInterface):
    """Create a Job object given the user session and an app id.

    Parameters
    ----------
    session: Session
        The user session
    appid: str
        The app ID
    execid: str
        The ID of the job if it has already been submited
    job_json : dict
        A dictionary from which to instantiate the job. Useful when creating a job from data from the API
    strict : bool
        If strict is True, Job creation will fail if not able to load correctly existing app configuration into the Job.
    """

    def __init__(
        self,
        session,
        appid,
        execid=None,
        job_json=None,
        strict=True,
    ):
        from playmolecule.datacenter import DataCenter

        super().__init__()
        self._session = session
        self._appid = appid
        self._execid = execid
        self._parentid = None

        self._datacenter = DataCenter(self._session)

        self._periodicity = 0
        self._arg("inputpath", "string", "input path", "/tmp", val.String())
        self._arg("output", "string", "Output path", ".", val.String())
        self._arg("description", "string", "job description", "", val.String())
        self._arg("group", "string", "job group", "", val.String())
        self._arg("name", "string", "job name", "", val.String())

        # Load basic job info for the above properties
        if job_json is not None:
            self._job_properties_from_json(job_json)

        # Create properties for the App-specific arguments
        if self._appid is not None:
            self._create_app_properties()
            if job_json is not None:
                # Load existing job configuration
                job_config_json = self._job_config_exists(job_json, strict)
                if job_config_json is not None:
                    self._job_config_from_json(job_config_json, strict)

    def _job_config_exists(self, job_json, strict):
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = self._download(
                "config", tmpdir, execid=job_json["id"], _logger=False
            )
            if config_file is None:
                msg = f"Job {job_json['id']} doesn't have a config object."
                if strict:
                    raise JobBucketNotFoundError(msg)
                else:
                    logger.warning(msg)
                    return None

            config_json = json.load(open(config_file, "rb"))
        return config_json

    def _create_app_properties(self):
        response = _request_url(
            "get",
            f"{self._server_endpoint}/apps/{self._appid}",
            headers={"token": self._token},
        )
        if response is None:
            raise RuntimeError("Could not get App information from server")

        app_arguments = json.loads(response.text)
        for param in app_arguments["params"]:
            validator = self._get_validator(param["type"])
            value = self._get_value(param["type"], param["value"], param["name"])
            self._arg(
                param["name"],
                param["type"],
                param["description"],
                value,
                validator,
                required=param["mandatory"],
            )

    def _job_properties_from_json(self, job_json):
        self._appid = job_json["appid"]
        self._execid = job_json["id"]
        self._parentid = job_json["parentid"]
        self.name = job_json["name"]
        self.group = job_json["group"]
        self.description = job_json["description"]

    def _job_config_from_json(self, job_config_json, strict):
        params = job_config_json["params"]
        for i in range(0, len(params)):
            key = params[i]["name"]
            value = params[i]["value"]
            try:
                value = self._get_value(params[i]["type"], value, key)
                setattr(self, key, value)
            except Exception as e:
                msg = f"Could not set Job attribute {key} with error {e}. The Job was possibly spawned with an older app version."
                if strict:
                    raise JobConfigNotLoadedError(msg)
                else:
                    logger.warning(msg)

    # TODO: accept more synonyms
    def _get_validator(self, type):
        if type == "bool":
            return val.Boolean()
        elif type == "int":
            return val.Number(int, "ANY")
        elif type == "string":
            return val.String()
        elif type == "file":
            return val.String()
        elif type == "float":
            return val.Number(float, "ANY")
        elif type == "fileORstring":
            return val.String()

    def _get_value(self, type, value, name):
        if value is None:
            return None

        if type == "bool":
            if value.lower() == "true":
                return True
            elif value.lower() == "false":
                return False
            raise ValueError(f'Invalid bool value "{value}" for argument {name}')
        elif type == "int":
            if value == "":
                return None
            try:
                converted_value = int(value)
                return converted_value
            except ValueError:
                raise ValueError(f'Invalid int value "{value}" for argument {name}')
        elif type == "float":
            if value == "":
                return None
            try:
                converted_value = float(value)
                return converted_value
            except ValueError:
                raise ValueError(f'Invalid float value "{value}" for argument {name}')
        elif type == "string" or type == "fileORstring" or type == "file":
            return value

    @property
    def _token(self):
        return self._session._token

    @property
    def _server_endpoint(self):
        return self._session._server_endpoint

    @property
    def _short_execid(self):
        return short_execid(self._execid)

    def submit(self, child_of=None, _logger=True):
        """Submits the job to the PlayMolecule backend server.

        Parameters
        ----------
        child_of: str
            The id of another job. If provided, the new job will be submited as a child of that job.
        _logger: bool
            Set to False to reduce verbosity
        """

        appid = self._appid
        token = self._token
        headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "token": token,
        }
        # send execution with Description
        description = self.description

        request_dict = {
            "description": description,
            "group": self.group,
            "name": self.name,
        }
        if child_of is not None:
            request_dict["parentid"] = child_of

        r = _request_url(
            "post",
            f"{self._server_endpoint}/apps/{appid}",
            headers=headers,
            json=request_dict,
            _logger=_logger,
        )
        if r is None:
            return

        if _logger:
            logger.debug(f"{self._server_endpoint}/apps/{appid}")
            logger.debug(r.text)

        python_obj = json.loads(r.text)
        config = python_obj["config"]
        token = python_obj["token"]
        self._execid = str(python_obj["id"])

        inputpath = os.path.join(self.inputpath, self._execid)
        os.makedirs(inputpath, exist_ok=True)

        # Check if files exist and copy them to the temp folder
        out_config = {
            "appid": config["appid"],
            "execid": self._execid,
            "params": [],
        }
        for param in config["params"]:
            name = param["name"]
            value = getattr(self, name)
            datatype = self._commands[name]["datatype"]
            if value is None and self._commands[name]["required"]:
                _throwError(f"Mandatory field '{name}' is empty.", _logger)
                return

            if (
                (datatype == "file" or datatype == "fileORstring")
                and value != ""
                and not value.startswith("dc://")
            ):
                value = getattr(self, name)
                if not os.path.isfile(value) and not os.path.isdir(value):
                    if (datatype == "file" or "." in value) and value is not None:
                        _throwError(f"File or folder does not exist: {value}", _logger)
                        return
                else:
                    filename = os.path.basename(os.path.normpath(value))
                    if os.path.isfile(value):
                        shutil.copy(
                            getattr(self, name),
                            os.path.join(inputpath, filename),
                        )
                    else:
                        shutil.copytree(
                            getattr(self, name),
                            os.path.join(inputpath, filename),
                        )
            else:
                if type(value) == bool:
                    value = str(value)

            out_config["params"].append(
                {
                    "name": name,
                    "value": value,
                    "type": param["type"],
                    "tag": param["tag"],
                }
            )

        # Write the current config out
        with tempfile.TemporaryDirectory() as tmpdir:
            config_file = os.path.join(tmpdir, f"{self._execid}.json")
            with open(config_file, "w") as f:
                json.dump(out_config, f, sort_keys=True, indent=4)
            self._upload("config", config_file, _logger=_logger)

        # Upload the inputs
        self._upload("input", inputpath)

        if os.path.isdir(inputpath):
            shutil.rmtree(inputpath, ignore_errors=True)

        # Send the start message for the execution
        request_dict = {"periodicity": self._periodicity}
        r = _request_url(
            "post",
            f"{self._server_endpoint}/jobs/{self._execid}",
            headers=headers,
            json=request_dict,
            _logger=_logger,
        )
        if r is None:
            return

        if _logger:
            logger.info(f"Execution submitted. Execution ID: {self._execid}")

    def _set_progress(self, info, completion, _logger=True):
        """Set the progress of a job execution.

        Parameters
        ----------
        info: str
            A string to be set as info progress
        completion: str
            The completion ratio of the job as a string in X/Y format where X current step, Y total steps.
        _logger: bool
            Set to False to reduce verbosity

        Examples
        --------
        >>> job.set_progress(info="Calculating features", completion="2/5")
        """
        if type(info) is not str or type(completion) is not str:
            _throwError("Error: 'info' and 'completion' must be strings", _logger)
            return

        token = self._token
        headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "admin": token,
        }
        request_dict = {"progressper": completion, "progressinfo": info}

        r = _request_url(
            "post",
            f"{self._server_endpoint}/jobs/{self._execid}/report",
            headers=headers,
            json=request_dict,
            _logger=_logger,
        )
        if r is None:
            return

    def retrieve(
        self,
        path=None,
        force=False,
        skip=False,
        on_status=(
            JobStatus.RUNNING,
            JobStatus.COMPLETED,
            JobStatus.ERROR,
            JobStatus.SLEEPING,
        ),
        attempts=5,
        execid_subdir=True,
        tmpdir=None,
        _logger=True,
    ):
        """Retrieve the results of a job execution.

        The output will be retrieved in the working directory if the argument path is not passed,
        inside a folder with the id job as a name.
        If this folder already exists, another one with a timestamp will be created.
        If the job has not completed, the function will try to download intermediate results.

        Parameters
        ----------
        path: str
            Destination folder
        force: bool
            If the destination folder exists and force == True, the destination folder will be removed.
        skip: bool
            If the destination folder exists and skip == True, nothing will be performed.
        on_status: list or JobStatus
            Restrict at which job status the results will be retrieved
        attempts: int
            How many times to attempt downloading the data from minio
        execid_subdir: bool
            If True it will create a subdirectory inside `path` with the execution ID as name and retrieve the results there
        tmpdir : str
            Set a different directory for storing temporary files. If set to None it defaults to /tmp/
        _logger: bool
            Set as True for printing the info and errors in sys.stdout. If False, it returns the same information as an object (default=True).

        Returns
        -------
        path: str
            The destination path of the retrieved data

        Examples
        --------
        >>> job.retrieve("./results/", on_status=(JobStatus.COMPLETED, JobStatus.ERROR))
        """
        on_status = ensurelist(on_status)

        path = os.path.abspath(self.output) if path is None else os.path.abspath(path)
        if execid_subdir:
            path = os.path.join(path, self._execid)
        path = get_destination_folder(path, force=force, skip=skip)
        if skip and os.path.exists(path):
            if _logger:
                logger.warning(f"Path already exists, skipping: {path}")
            return path

        if not hasattr(self, "_execid"):
            _throwError("Can't retrieve. Job does not exist.", _logger)
            return

        if _logger:
            logger.info("Retrieving data of execution: " + self._execid)

        job_info = self._getJobInfo(_logger=_logger)
        if job_info is None:
            return

        status = JobStatus(int(job_info["status"]))

        if status in on_status:
            self._download(
                "output",
                path,
                tmpdir=tmpdir,
                attempts=attempts,
                _logger=_logger,
            )
            return path
        else:
            if _logger:
                logger.info(
                    f"Current job status is '{status.describe()}' but only retrieving '{', '.join(map(str, on_status))}'"
                )

    def describe(self, _logger=True):
        """Describes the input parameters of an application

        Parameters
        ----------
        _logger: bool
            Set as True for printing the info and errors in sys.stdout.
            If False, it returns the same information as a list (default=True).

        Returns
        -------
        app_descr : list
            The description of the inputs of an app
        """
        from playmolecule.session import _describe_app

        r = _request_url(
            "get",
            f"{self._server_endpoint}/apps/{self._appid}",
            headers={"token": self._token},
            _logger=_logger,
        )
        if r is None:
            return

        python_obj = json.loads(r.text)
        if _logger:
            _describe_app(python_obj, self)
        else:
            return python_obj

    def get_status(self, _logger=True):
        """Prints or returns the job status

        Parameters
        ----------
        _logger: bool
            Set as True for printing the info and errors in sys.stdout.
            If False, it returns the same information (default=True).

        Returns
        -------
        status : JobStatus
            The current status of the job
        """
        job_info = self._getJobInfo(_logger=_logger)
        if job_info is None:
            return

        status = JobStatus(job_info["status"])

        if _logger:
            logger.info(f"Job Status: {status.describe()}")

        return status

    def get_progress(self, _logger=True):
        """Prints the job progress.

        Parameters
        ----------
        _logger: bool
            Set as True for printing the info and errors in sys.stdout. If False, it returns the same information as an object (default=True).

        Returns
        -------
        info : str
            The information on the current progress
        completion : str
            The completion percentage
        """
        info = ""
        completion = ""
        if hasattr(self, "_execid"):
            job_info = self._getJobInfo(_logger=_logger)
            if job_info is None:
                return

            info = job_info["progressinfo"]
            completion = job_info["progressper"]
            if _logger:
                print("%-20s %-20s" % ("Progress info:", info))
                print("%-20s %-20s" % ("Progress percentage:", completion))
            else:
                return info, completion
        else:
            _throwError("Job not submited", _logger)
            return

    def _set_start(self, _logger=True):
        """Sets the status of a job as running

        Parameters
        ----------
        _logger: bool
            Set to False to reduce verbosity
        """
        self._session.set_job_status(self._execid, JobStatus.RUNNING, _logger=_logger)

    def _set_complete(self, _logger=True):
        """Sets the status of a job as completed

        Parameters
        ----------
        _logger: bool
            Set to False to reduce verbosity
        """
        self._session.set_job_status(self._execid, JobStatus.COMPLETED, _logger=_logger)

    def _set_sleep(self, _logger=True):
        """Sets the status of a job as sleeping

        Parameters
        ----------
        _logger: bool
            Set to False to reduce verbosity.
        """
        self._session.set_job_status(self._execid, JobStatus.SLEEPING, _logger=_logger)

    def _set_error(self, error_info=None, _logger=True):
        """Sets the status of a job as error

        Parameters
        ----------
        error_info : str
            Pass verbose information on the cause of the error
        _logger: bool
            Set to False to reduce verbosity
        """
        self._session.set_job_status(
            self._execid, JobStatus.ERROR, error_info=error_info, _logger=_logger
        )

    def get_children(self, status=None, return_dict=False, _logger=True):
        """Returns the children of the job.

        Parameters
        ----------
        status: list of JobStatus
            Only get children with specific job statuses given in a list
        return_dict : bool
            Set to True to return a dictionary with the children information instead of Job objects
        _logger : bool
            Set to False to reduce verbosity

        Returns
        -------
        children : list
            The children of the job
        """
        if not hasattr(self, "_execid"):
            return None

        token = self._token
        headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "token": token,
        }
        if status is None:
            r = _request_url(
                "get",
                f"{self._server_endpoint}/jobs/{self._execid}/children",
                headers=headers,
                _logger=_logger,
            )
        else:
            request_dict = {"statuscodes": [int(s) for s in ensurelist(status)]}
            r = _request_url(
                "post",
                f"{self._server_endpoint}/jobs/{self._execid}/children",
                headers=headers,
                json=request_dict,
                _logger=_logger,
            )

        if r is None:
            return None

        python_obj = json.loads(r.text)
        if return_dict:
            return python_obj
        else:  # This will become the default
            childrenJobs = []
            statuses = []
            for child_json in python_obj:
                childrenJobs.append(
                    Job(session=self._session, appid=None, job_json=child_json)
                )
                statuses.append(child_json["status"])
            return childrenJobs, statuses

    def wait_children(
        self,
        on_status=(JobStatus.COMPLETED, JobStatus.ERROR),
        seconds=10,
        _logger=True,
    ):
        """Blocks execution until all the children of the job have reached the given status

        Parameters
        ----------
        on_status : JobStatus or list of JobStatus
            The status(es) at which a child job will stop being waited upon
        seconds : float
            The sleep time between status checking of the children jobs
        _logger: bool
            Set to False to reduce verbosity
        """
        from collections import Counter

        unfinishedStatus = set(JobStatus) - set(ensurelist(on_status))

        while True:
            running_child_jobs = self.get_children(
                status=unfinishedStatus, return_dict=True
            )
            nrunning = len(running_child_jobs)

            if nrunning == 0:
                break

            statuses = Counter(
                [str(JobStatus(cj["status"])) for cj in running_child_jobs]
            )

            if _logger:
                logger.info(
                    f"There are still {nrunning} children running with status {dict(statuses)}. Sleeping for {seconds} seconds."
                )
            time.sleep(seconds)

        if _logger:
            logger.info("No children currently running")

    def wait(
        self,
        on_status=(JobStatus.COMPLETED, JobStatus.ERROR),
        seconds=10,
        _logger=True,
    ):
        """Blocks execution until the job has reached the specific status or any of a list of statuses.

        Parameters
        ----------
        on_status : JobStatus or list of JobStatus
            The status(es) at which the job will not be waited upon and the code execution will continue.
        seconds : float
            The sleep time between status cheching
        _logger: bool
            Set to False to reduce verbosity
        """
        on_status = ensurelist(on_status)

        while True:
            status = self.get_status(_logger=False)
            if status in on_status:
                break

            if _logger:
                logger.info(
                    f"Job status \"{status.describe()}\". Waiting for it to reach ({', '.join(map(str, on_status))}). Sleeping for {seconds} seconds."
                )
            time.sleep(seconds)

    def read_arguments(self, filename, ftype="yaml"):
        """Reads argument values of the job from a json or yaml file

        Parameters
        ----------
        filename : str
            The name of the file (incl. extension) from which to read the arguments
        ftype : str
            Can be either "json" or "yaml" depending on what file format you want to read
        """
        import yaml

        with open(filename, "r") as f:
            if ftype == "json":
                argdict = json.load(f)
            elif ftype == "yaml":
                argdict = yaml.load(f, Loader=yaml.FullLoader)
            else:
                raise AttributeError(
                    f"Invalid file type {ftype}. Can be only 'json' or 'yaml'."
                )

        for key, value in argdict.items():
            setattr(self, key, value)

    def write_arguments(self, filename, ftype="yaml"):
        """Writes the current argument values of the job to a json or yaml file

        Parameters
        ----------
        filename : str
            The name of the file (incl. extension) in which to write the arguments
        ftype : str
            Can be either "json" or "yaml" depending on what file format you want to write
        """
        import yaml

        argdict = {
            key: value for key, value in self.__dict__.items() if key in self._commands
        }
        with open(filename, "w") as f:
            if ftype == "json":
                json.dump(argdict, f, sort_keys=True, indent=4)
            elif ftype == "yaml":
                yaml.dump(argdict, f)
            else:
                raise AttributeError(
                    f"Invalid file type {ftype}. Can be only 'json' or 'yaml'."
                )

    def cancel(self, _logger=True):
        """Cancel the job

        Cancels the job. This will set the status to ERROR and stop any running execution (children as well)
        """
        self._session.cancel_job(self._execid, _logger=_logger)

    def _delete(self, _logger=True):
        """Deletes the job and all it's information from the backend"""
        self._session.delete_job(self._execid, _logger=_logger)

    def _getJobInfo(self, execid=None, _logger=True):
        if execid is None:
            if not hasattr(self, "_execid"):
                _throwError(
                    "No execid found in job object. The job has possibly not been submitted yet.",
                    _logger,
                )
            execid = self._execid

        headers = {
            "Content-type": "application/json",
            "Accept": "text/plain",
            "token": self._token,
        }

        r = _request_url(
            "get",
            f"{self._server_endpoint}/jobs/{execid}",
            headers=headers,
            _logger=_logger,
        )
        return json.loads(r.text) if r is not None else None

    def get_error_info(self, _logger=True):
        """Get information related to an errored job

        Parameters
        ----------
        _logger: bool
            Set as True for printing the info and errors in sys.stdout.
            If False, it returns the same information (default=True).

        Returns
        -------
        error_info : str
            Information on the cause of the error
        """
        job_info = self._getJobInfo(_logger=_logger)
        if job_info is None:
            return ""

        error_info = job_info.get("error_info", "")

        if _logger:
            logger.info(f"Error Info: {error_info}")

        return error_info

    def _download(self, name, path, tmpdir=None, attempts=5, execid=None, _logger=True):
        if name not in ("input", "output", "config", "intermediate"):
            raise RuntimeError(
                "name argument must be either input, output, config or intermediate"
            )

        if execid is None:
            execid = self._execid

        ds = self._datacenter.get_datasets(remotepath=f"{execid}/{name}", _logger=False)
        if len(ds) == 0:
            if _logger:
                logger.info(f"No data found for dataset {execid}/{name}")
            return

        assert (
            len(ds) == 1
        ), "More than one job dataset found with the same name. This should never happen."

        return self._datacenter.download_dataset(
            ds[0]["id"], path, tmpdir=tmpdir, attempts=attempts, _logger=_logger
        )

    def _upload(
        self,
        name,
        path,
        comments=None,
        public=False,
        overwrite=True,
        tmpdir=None,
        attempts=5,
        execid=None,
        _logger=True,
    ):
        if name not in ("input", "output", "config", "intermediate"):
            raise RuntimeError(
                "name argument must be either input, output, config or intermediate"
            )

        if execid is None:
            execid = self._execid

        appname = self._appid.lower().split("-")[0]
        if comments is None:
            try:
                comments = f"{self.name} {self.description} {self.group}".strip()
            except Exception:
                comments = ""

        self._datacenter.upload_dataset(
            localpath=path,
            remotepath=f"{execid}/{name}",
            comments=comments,
            public=public,
            execid=execid,
            overwrite=overwrite,
            tmpdir=tmpdir,
            attempts=attempts,
            _logger=_logger,
            tags=[f"app:{appname}:{name}"],
        )



from configparser import ConfigParser
import playmolecule
import inspect
import logging
import os


logger = logging.getLogger(__name__)


_config = ConfigParser()


def loadConfig():
    if os.getenv("PM_SDK_CONFIG"):
        path = os.environ["PM_SDK_CONFIG"]
    elif os.getenv("SDKCONFIG"):
        path = os.environ["SDKCONFIG"]
    else:
        homeDir = os.path.dirname(inspect.getfile(playmolecule))
        path = os.path.join(homeDir, "config.ini")

    logger.info(f"Reading PM API configuration from {path}")
    _config.read(path)



from playmolecule.utils import (
    tempname,
    format_filesize,
    parse_timestamp_utc,
    utc_timestamp_to_local,
    ensurelist,
    _requestURL,
)
import tarfile
import logging
import time
import json
import os
import re


logger = logging.getLogger(__name__)


class DataCenterError(Exception):
    pass


def _throw_error(message, _logger=False):
    if _logger:
        logger.error(message)
    else:
        raise DataCenterError(message)


def _request_url(*args, **kwargs):
    return _requestURL(*args, **kwargs, _throwError=_throw_error)


class DataCenter:
    """Class which manages all the datasets in the PlayMolecule backend

    Parameters
    ----------
    session : Session object
        A Session object
    """

    def __init__(self, session):
        self._session = session

    def get_datasets(
        self,
        public=None,
        datasetid=None,
        remotepath=None,
        useronly=False,
        startswith=None,
        tags=None,
        taggedonly=False,
        completedonly=False,
        _logger=True,
    ):
        """Get a list of datasets filtered by various arguments

        Parameters
        ----------
        public : bool
            If set to True it will only return public datasets. If set to False it will only return private datasets.
            If set to None this parameter will be ignored.
        datasetid : int
            The ID of a specific dataset for which we want to retrieve information
        remotepath : str
            The remote (virtual) path at which the dataset is located
        useronly : bool
            Returns only datasets of the currently logged in user of the Session
        startswith : str
            Returns any datasets whose remote (virtual) path starts with the specific string
        tags : list
            Returns only datasets which have the specified tags
        taggedonly : bool
            If set to True it will only return datasets which have tags
        completedonly : bool
            If set to True it will only return datasets whose jobs have completed successfully
        _logger : bool
            Set to False to reduce verbosity

        Returns
        -------
        datasetlist : list
            A list of datasets retrieved with the above filters

        Examples
        --------
        >>> datasets = dc.get_datasets(remotepath="KdeepTrainer/models/PDBBind2019")
        >>> datasets = dc.get_datasets(tags=["app:kdeep"])
        """
        self._session._require_log_in()

        if tags is not None:
            tags = ensurelist(tags)

        headers = {
            "token": self._session._token,
            "userOnly": str(useronly),
            "datasetId": str(datasetid) if datasetid is not None else "",
            "filePath": remotepath if remotepath is not None else "",
            "public": str(public) if public is not None else "",
            "startsWith": startswith if startswith is not None else "",
            "tags": ",".join(tags) if tags is not None else "",
            "taggedOnly": str(taggedonly),
            "completedOnly": str(completedonly),
        }

        r = _request_url(
            "get",
            f"{self._session._server_endpoint}/datacenter",
            headers=headers,
            _logger=False,
        )
        if r is None:
            return

        datasets = json.loads(r.text)
        for ds in datasets:
            ds["created_at"] = parse_timestamp_utc(ds["created_at"])
            ds["updated_at"] = parse_timestamp_utc(ds["updated_at"])

        if _logger:
            columns = "{:<10} {:<49} {:<20} {:<6} {:<8} {:<36} {:<20} {:<20}"
            print(
                columns.format(
                    "ID",
                    "RemotePath",
                    "Comments",
                    "Public",
                    "FileSize",
                    "ExecID",
                    "DateCreated",
                    "DateUpdated",
                )
            )

            for ds in datasets:
                print(
                    columns.format(
                        ds["id"],
                        ds["filepath"],
                        ds["comments"][:20],
                        "True" if ds["public"] else "False",
                        format_filesize(ds["filesize"]),
                        ds["execid"],
                        utc_timestamp_to_local(ds["created_at"]).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                        utc_timestamp_to_local(ds["updated_at"]).strftime(
                            "%Y-%m-%d %H:%M:%S"
                        ),
                    )
                )
        else:
            return [] if datasets is None else datasets

    def upload_dataset(
        self,
        localpath,
        remotepath,
        comments="",
        public=False,
        execid="",
        overwrite=False,
        tags=None,
        tmpdir=None,
        attempts=1,
        _logger=True,
    ):
        """Uploads a dataset to the backend data center

        Parameters
        ----------
        localpath : str
            The location of the file we want to upload
        remotepath : str
            The remote (virtual) location to which the file should be uploaded
        comments : str
            Comments to attach to the specific dataset
        public : bool
            Set to True to make the dataset public (available to all users)
        execid : str
            Optionally you can relate this dataset to a specific job execution by passing it's job ID
        overwrite : bool
            Set to True to overwrite existing datasets at the specified remote (virtual) location
        tags : list of str
            A list of tags to attach to the specific dataset
        tmpdir : str
            Location to use for creating the upload archive file. The file will be deleted after uploading.
            If set to None it will use /tmp/
        attempts : int
            Number of times to attempt uploading the file. Can help with unstable connections
        _logger : bool
            Set to False to reduce the verbosity
        """
        self._session._require_log_in()

        if not os.path.isfile(localpath) and not os.path.isdir(localpath):
            raise RuntimeError("upload_dataset only accepts files or folders")

        if tags is not None:
            tags = ensurelist(tags)

        fname = os.path.basename(os.path.abspath(localpath))

        tmpfile = tempname(suffix=".tar.gz", tmpdir=tmpdir)
        with tarfile.open(tmpfile, "w:gz") as tar:
            if os.path.isdir(localpath):
                tar.add(localpath, arcname="")
            else:
                tar.add(localpath, arcname=fname)

        # If you specify a remote folder, append the file name for the remote path
        if remotepath.endswith("/"):
            remotepath = remotepath + fname

        # Tags need to be all lower case and strip whitespace
        if tags is not None:
            tags = [tag.lower().strip().replace(" ", "_") for tag in tags]
            for tag in tags:
                if not re.match("^[a-zA-Z0-9_:]+$", tag):
                    raise RuntimeError(
                        "Tags need to be alphanumeric only and can contain _ or :"
                    )

        files = {"uploadfile": (fname, open(tmpfile, "rb"))}
        headers = {"token": self._session._token}
        data = {
            "remotepath": remotepath,
            "comments": comments,
            "public": public,
            "execid": execid,
            "overwrite": overwrite,
            "tags": ",".join(tags) if tags is not None else "",
        }

        uploaded = False
        while not uploaded and attempts > 0:
            try:
                result = _request_url(
                    "post",
                    f"{self._session._server_endpoint}/datacenter",
                    headers=headers,
                    files=files,
                    data=data,
                    _logger=_logger,
                )
            except Exception as e:
                attempts -= 1
                _throw_error(
                    f"Error uploading dataset {localpath} with error {e}. Retrying in 5s. Attempts remaining {attempts}.",
                    _logger,
                )
                time.sleep(5)
            else:
                uploaded = True

        os.remove(tmpfile)

        if not uploaded or result is None:
            _throw_error(
                f"Failed to upload dataset {localpath} to {remotepath}",
                _logger,
            )
            return

        if _logger:
            logger.info(f"File {localpath} was uploaded successfully.")

        return int(result.json()["id"])

    def remove_dataset(self, datasetid, _logger=True):
        """Removes (deletes) a dataset from the backend

        Parameters
        ----------
        datasetid : int
            The ID of the dataset we want to delete
        _logger : bool
            Set to False to reduce verbosity
        """
        self._session._require_log_in()

        headers = {"token": self._session._token}

        r = _request_url(
            "delete",
            f"{self._session._server_endpoint}/datacenter/{datasetid}",
            headers=headers,
            _logger=_logger,
        )

        if r is None:
            return

        if _logger:
            logger.info(f"Dataset {datasetid} was deleted successfully.")

    def download_dataset(self, datasetid, path, tmpdir=None, attempts=1, _logger=True):
        """Download a dataset from the backend

        Parameters
        ----------
        datasetid : int
            The ID of the dataset we want to download
        path : str
            The location to which to download the dataset
        tmpdir : str
            A location to store temporary data. If set to None, the default is /tmp/
        attempts : int
            Number of times to attempt uploading the file. Can help with unstable connections
        _logger : bool
            Set to False to reduce the verbosity

        Examples
        --------
        >>> dc.download_dataset(182, "./dataset_182/")
        """
        self._session._require_log_in()

        if not isinstance(datasetid, int):
            raise RuntimeError("datasetid must be an integer")

        downloaded = False
        while not downloaded and attempts > 0:
            try:
                result = _request_url(
                    "get",
                    f"{self._session._server_endpoint}/datacenter/{datasetid}",
                    headers={"token": self._session._token},
                    _logger=_logger,
                )
            except Exception as e:
                attempts -= 1
                _throw_error(
                    f"Error downloading dataset {datasetid} with error {e}. Retrying in 5s. Attempts remaining {attempts}.",
                    _logger,
                )
                time.sleep(5)
            else:
                downloaded = True

        if not downloaded or result is None:
            _throw_error(
                f"Failed to download dataset {datasetid} to {path}",
                _logger,
            )
            return

        tmpfile = tempname(suffix=".tar.gz", tmpdir=tmpdir)

        with open(tmpfile, "wb") as f:
            f.write(result.content)

        os.makedirs(path, exist_ok=True)

        try:
            with tarfile.open(tmpfile) as tar:
                member = tar.getmembers()[0]
                outname = os.path.abspath(os.path.join(path, member.name))
                tar.extractall(path=path)
            os.remove(tmpfile)
        except Exception as e:
            if os.path.exists(tmpfile):
                os.remove(tmpfile)
            _throw_error(
                f"Error untaring results for dataset {datasetid} with error {e}",
                _logger,
            )
            return

        if _logger:
            logger.info(
                f"Dataset {datasetid} was downloaded successfully to {outname}."
            )

        return outname

    def get_dataset_tags(self, datasetid, _logger=True):
        """Returns tags associated with a dataset

        Parameters
        ----------
        datasetid : int
            The ID of the dataset for which to return tags
        _logger : bool
            Set to False to reduce verbosity
        """
        self._session._require_log_in()

        headers = {"token": self._session._token}

        r = _request_url(
            "get",
            f"{self._session._server_endpoint}/datacenter/{datasetid}/tags",
            headers=headers,
            _logger=_logger,
        )

        if r is None:
            return

        response = json.loads(r.text)
        return [] if response is None else response

    def remove_dataset_tag(self, datasetid, tag, _logger=True):
        """Removes a tag attached to a dataset

        Parameters
        ----------
        datasetid : int
            The ID of the dataset
        tag : str
            The tag we want to remove from the dataset
        _logger : bool
            Set to False to reduce verbosity
        """
        self._session._require_log_in()

        headers = {"token": self._session._token}

        r = _request_url(
            "delete",
            f"{self._session._server_endpoint}/datacenter/{datasetid}/tags/{tag.lower()}",
            headers=headers,
            _logger=_logger,
        )

        if r is None:
            return

        if _logger:
            logger.info(f"Tag {tag} was deleted from dataset {datasetid} successfully.")



# Utilities necessary inside PM apps
import os


def tarFolder(folder, outtar=None, arcname=""):
    import tarfile

    folder = os.path.abspath(folder)
    parentdir = os.path.dirname(folder)
    foldername = os.path.basename(folder)

    if outtar is None:
        outtar = os.path.join(parentdir, f"{foldername}.tar.gz")

    with tarfile.open(outtar, "w:gz") as tar:
        tar.add(folder, arcname=arcname)

    return outtar


def getJobId(configpath):
    """
    Returns the job id stored in the config file

    Parameters
    ----------
    configpath: str
        The file path for the config file

    Returns
    -------
    jobid: str
        The id of the job
    """
    import json

    jobId = json.load(open(configpath, "r"))["execid"]
    return jobId


def setAsCompleted(outdir):
    """
    Writes the sentinel for the complete job.

    Parameters
    ----------
    outdir: str
        The location where to write the sentinel file
    """
    writeSentinel("pmwsjob.done", outdir)


def setAsError(outdir):
    """
    Writes the sentinel for the error job.

    Parameters
    ----------
    outdir: str
        The location where to write the sentinel file
    """
    writeSentinel("pmwsjob.err", outdir)


def setAsSleep(outdir):
    """
    Writes the sentinel for the complete job.

    Parameters
    ----------
    outdir: str
        The location where to write the sentinel file
    """
    writeSentinel("pmwsjob.sleep", outdir)


def writeSentinel(fname, outfolder, message=None):
    """
    Writes a sentinel file. A message can be written in the file

    Parameters
    ----------
    fname: str
        The sentinel file name
    outfolder: str
        The location where to write the sentinel file
    message: str
        The text to write in the sentinel file

    """
    import yaml

    fout = os.path.join(outfolder, fname)

    f = open(fout, "w")
    if message is not None:
        if isinstance(message, dict):
            yaml.dump(message, f, default_flow_style=True)
        else:
            f.write("{}\n".format(message))
    f.close()


def basenameInputs(parser, args, prefix="."):
    """
    Returns the basename of the value for the file arguments

    Parameters
    ----------
    parser: argparse.ArgumentParser
        The argparse parser
    args: dict
        The dictionary of the arguments to inspect
    prefix: str
        The path prefix to apply to the basename
    Returns
    -------
    args: dict
        The dictionary of the arguments with the basename value (for files)
    """
    for action in parser._actions:
        if (
            action.metavar == "FILE"
        ):  # For DIRECTORY should not use basename but the equivalent
            value = args[action.dest]
            if value is None or value == "":
                continue
            if isinstance(value, list):
                tmp_args = []
                for v in value:
                    if v == "":
                        continue
                    bname = os.path.basename(v)
                    oname = os.path.join(prefix, bname)
                    tmp_args.append(oname)
                args[action.dest] = tmp_args
            else:
                bname = os.path.basename(value)
                tmponame = (
                    bname if bname != "" else os.path.normpath(value).split("/")[-1]
                )
                oname = os.path.join(prefix, tmponame)
                args[action.dest] = oname
    return args


def writeInputsOptionsLog(path, parser, varset, debug=False):
    """
    Writes a log file, in yml format file, reporting all the options used for the current simulations.

    Parameters
    ----------
    parser: argparse.ArgumentParser
        The argparse parser
    path: str
        The folder destination where to write the log
    varset: Namespace
        The namespace that contains all the options used
    debug: bool
        Set as True if you want to write the debug value
    """
    import yaml

    fname = os.path.join(path, "inopt.yml")
    options = {k: ("" if v is None else v) for k, v in vars(varset).items()}
    if options["outdir"].startswith("/data/out"):
        del options["outdir"]
    if options["scratchdir"].startswith("/data/scratch"):
        del options["scratchdir"]
    if not debug:
        del options["debug"]
    options = basenameInputs(parser, options, prefix="")

    with open(fname, "w") as outfile:
        yaml.dump(options, outfile, default_flow_style=False)


def zipdir(path, ziph):
    """
    Adds all the tree files of the folder passed into the zip file
    Parameters
    ----------
    path: str
        The path to add to the zip file
    ziph: zipfile.ZipFile
        The zipfile object
    """
    # ziph is zipfile handle
    for root, _, files in os.walk(path):
        for file in files:
            ziph.write(
                os.path.join(root, file),
                os.path.relpath(os.path.join(root, file), os.path.join(path, "..")),
            )


def zipit(dir_list, zip_name):
    """
    Creates the zip file with all the files and folder passed

    Parameters
    ----------
    dir_list: list
        The list of files and folders to zip
    zip_name: str
        The name of the zipfile
    """
    import zipfile

    zipf = zipfile.ZipFile(zip_name, "w", zipfile.ZIP_DEFLATED)
    for dir in dir_list:
        if os.path.isdir(dir):
            zipdir(dir, zipf)
        else:
            zipf.write(dir, os.path.basename(dir))
    zipf.close()


def unzipit(z, tmpdir="./tmp", overwrite=False, outdir=None):
    """
    Returns a list with the files in the zip file (when a zip). In alternative, a list with the passed file/folder

    Paramteres
    ----------
    z: str
        The path file

    Returns
    -------
    inps: list
        A list of the file in zip or the file/folder as unique item list
    """
    import zipfile
    import shutil
    from glob import glob

    if not zipfile.is_zipfile(z):
        return [z]
    tmpdir = os.path.normpath(tmpdir)
    os.makedirs(tmpdir, exist_ok=True)

    with zipfile.ZipFile(z, "r") as zip_file:
        zip_file.extractall(tmpdir)

    tmpinps = glob(tmpdir + "/*")
    if outdir is None:
        outdir = os.getcwd()

    inps = []
    for f in tmpinps:
        b = os.path.basename(f)
        o = os.path.join(outdir, b)
        if os.path.exists(o) and not overwrite:
            continue
        if os.path.exists(o) and overwrite and os.path.isdir(o):
            shutil.rmtree(o)
        if os.path.exists(o) and overwrite and os.path.isfile(o):
            os.remove(o)
        shutil.move(f, o)
        inps.append(o)
    shutil.rmtree(tmpdir)

    return inps


def pmtryexcept(arg1, arg2, arg3):
    """
    The decorator that handles the try/except by redirecting the sys.stderr and sys.stdout to a log file.
    Two boolean arguments can trigger the usage of the log file: arg1 refers to standalone mode; arg2 refers to
    debug mode. For writing to the log file the arg1 needs to be False or the arg2 needs to be True.

    """
    import sys
    import traceback

    def wrap(f):
        def wrapped_f(self, *args, **kwargs):
            standalone = getattr(self, arg1)
            debug = getattr(self, arg2)
            pmwsjob = getattr(self, arg3)

            try:
                return f(self, *args, **kwargs)
            except Exception as e:
                err_detail = "\nError detailed:\n\t{}: {}\n".format(type(e).__name__, e)
                tb = repr(traceback.extract_tb(e.__traceback__, limit=-1))
                _fr = tb.split()
                _file = _fr[_fr.index("file") + 1]
                _line = _fr[_fr.index("line") + 1]
                eType, eValue, eTraceback = sys.exc_info()
                summary_error = (
                    "Error raised from:\n\tfile: {}\n\tline: {}\n\ttype: {}\n\tvalue: {}\n\t"
                    "traceback {}\n".format(_file, _line, eType, eValue, eTraceback)
                )

                log = "" if standalone else "LOG"
                h1 = "# ERROR {}\n".format(log)
                h2 = "\n# Error summary (dev)\n"
                h3 = "\n# Error Traceback (dev)\n"
                spacer = "_" * 40

                if standalone and not debug:
                    text = [h2, summary_error, spacer, h3]
                    print("".join(text))
                    traceback.print_tb(e.__traceback__)
                    text = ["\n", spacer, "\n", h1, err_detail]
                    print("".join(text))
                else:
                    error_log = (
                        "/data/out/outerror.log" if not debug else "outerror.log"
                    )
                    flog = open(error_log, "w")
                    text = [h1, err_detail, spacer, h2, summary_error, spacer, h3]
                    flog.write("".join(text))
                    traceback.print_tb(e.__traceback__, file=flog)

                    flog.close()
                    if pmwsjob is not None:
                        setAsError("/data/out/")
                        # pmwsjob.setError()
                    print("Error !!!")
                sys.exit(0)

        return wrapped_f

    return wrap



from playmolecule.session import (
    Session,
    SessionError,
    UserNotFoundError,
    UserFailedRegistrationError,
    UserUpdateError,
)
from playmolecule.job import (
    Job,
    JobStatus,
    JobBucketNotFoundError,
    JobConfigNotLoadedError,
)
from playmolecule.datacenter import DataCenter, DataCenterError
from playmolecule.jsonlogger import enable_json_logger
from playmolecule.config import loadConfig
import logging.config
import os


__init__dir = os.path.dirname(os.path.abspath(__file__))

try:
    logging.config.fileConfig(
        os.path.join(__init__dir, "logging.ini"), disable_existing_loggers=False
    )
except Exception as e:
    print(f"PlayMolecule: Logging setup failed with error {e}")

loadConfig()
